{% if entries %}
    <ul>
    {% for entry in entries %}
        <li>
            {{ entry.key }} 
            : 
            {{ entry.val }}
        </li>
    {% endfor %}
    </ul>
{% else %}
    <p>No entries are available.</p>
{% endif %}